from unittest import mock

import pytest

from horkos.critiquer import _casing


def test_schema_uniform_field_casing_check_with_critique():
    """Should identify mixed casing conventions within schema"""
    schema = mock.MagicMock()
    schema.fields = {
        'fooBar': 'something',
        'fiz_buzz': 'something',
    }

    result = _casing.schema_uniform_field_casing_check(schema)

    assert result
    expected = 'camelCase and snake_case are most common with 1 matching.'
    assert expected in result[0].message


def test_schema_uniform_field_casing_check_without_critique():
    """Should return none if the casing is consistent"""
    schema = mock.MagicMock()
    schema.fields = {
        'foobar': 'something',
        'fiz_buzz': 'something',
    }

    result = _casing.schema_uniform_field_casing_check(schema)

    assert not result


def test_relative_field_casing_check_with_critique():
    """Should identify mixed field conventions between schema and catalog."""
    catalog = mock.MagicMock(
        schemas={
            'foo': mock.MagicMock(fields={'fooBar': None, 'fiz_buzz': None}),
            'baz': mock.MagicMock(
                fields={'fooBar': None, 'boo_far': None, 'biz_fuzz': None}
            ),
        }
    )
    schema = mock.MagicMock(fields={'fooBar': None, 'bizFuzz': None})

    result = _casing.relative_field_casing_check(schema, catalog)

    assert result
    expected = (
        'The casing of bizFuzz does not match the dominant patterns of the '
        'catalog'
    )
    assert expected in result[0].message


def test_relative_field_casing_check_with_schema_in_catalog():
    """If the schema is already in the catalog an error should be raised."""
    catalog = mock.MagicMock(schemas={'foo': 'bar'})
    schema = mock.MagicMock()
    schema.name = 'foo'

    with pytest.raises(ValueError):
        _casing.relative_field_casing_check(schema, catalog)


def test_relative_field_casing_without_critique():
    """Should return None if their are no issues."""
    catalog = mock.MagicMock(
        schemas={
            'foo': mock.MagicMock(fields={'fooBar': None, 'fizBuzz': None}),
            'baz': mock.MagicMock(
                fields={'fooBar': None, 'booFar': None, 'bizFuzz': None}
            ),
        }
    )
    schema = mock.MagicMock(fields={'fooBar': None, 'bizFuzz': None})

    result = _casing.relative_field_casing_check(schema, catalog)

    assert not result


def test_catalog_field_casing_check_with_critique():
    """Should be able to identify inconsistent naming in a catalog of data."""
    catalog = mock.MagicMock(
        schemas={
            'foo': mock.MagicMock(fields={'fooBar': None, 'fiz_buzz': None}),
            'baz': mock.MagicMock(
                fields={'fooBar': None, 'boo_far': None, 'biz_fuzz': None}
            ),
        }
    )

    result = _casing.catalog_field_casing_check(catalog)

    assert result
    expected = (
        'No clear casing convention. snake_case is most '
        'common with 60% matching.'
    )
    assert expected in result[0].message


def test_catalog_field_casing_check_without_critique():
    """Should return None if all fields are consistent."""
    catalog = mock.MagicMock(
        schemas={
            'foo': mock.MagicMock(fields={'foobar': None, 'fiz_buzz': None}),
            'baz': mock.MagicMock(
                fields={'foobar': None, 'boo_far': None, 'biz_fuzz': None}
            ),
        }
    )

    result = _casing.catalog_field_casing_check(catalog)

    assert not result


def test_schema_prefer_snake_case_with_critique():
    """
    Should be able to identify when snake_case isn't used in a schema
    and recommend its usage.
    """
    schema = mock.MagicMock(fields={'fooBar': None, 'bizFuzz': None})

    result = _casing.schema_prefer_snake_case(schema)

    assert result
    expected = 'snake_case field names are recommended'
    assert expected in result[0].message


def test_schema_prefer_snake_case_without_critique():
    """If snake_case is already used, then return None."""
    schema = mock.MagicMock(fields={'foobar': None, 'fiz_buzz': None})

    result = _casing.schema_prefer_snake_case(schema)

    assert not result


def test_catalog_prefer_snake_case_with_critique():
    """
    Should be able to identify when snake_case isn't used in a schema
    and recommend its usage.
    """
    catalog = mock.MagicMock(
        schemas={
            'foo': mock.MagicMock(fields={'Foobar': None, 'fiz_buzz': None}),
            'baz': mock.MagicMock(
                fields={'Foobar': None, 'BooFar': None, 'biz_fuzz': None}
            ),
        }
    )

    result = _casing.catalog_prefer_snake_case(catalog)

    assert result
    expected = 'snake_case field names are recommended'
    assert expected in result[0].message


def test_catalog_prefer_snake_case_without_critique():
    """If snake_case is already used, then return None."""
    catalog = mock.MagicMock(
        schemas={
            'foo': mock.MagicMock(fields={'foobar': None, 'fiz_buzz': None}),
            'baz': mock.MagicMock(
                fields={'foobar': None, 'boo_far': None, 'biz_fuzz': None}
            ),
        }
    )

    result = _casing.catalog_prefer_snake_case(catalog)

    assert not result



import io
from unittest import mock

import pytest

from horkos.cmdline import critique


def test_expand_directories_with_nonexistent_file():
    """Should get a SystemExit if a given file doesn't exist."""
    output = io.StringIO()

    with pytest.raises(SystemExit):
        with mock.patch('sys.stderr', new=output):
            critique._expand_directories(['not-a-real-file.yaml'])

    result = output.getvalue()
    assert 'ERROR: not-a-real-file.yaml does not exist.' in result


@mock.patch('os.path.exists', new=lambda x: True)
@mock.patch('os.path.isdir', new=lambda x: False)
def test_expand_directories_with_nonschema_file():
    """Should get a SystemExit if a given file doesn't exist."""
    output = io.StringIO()

    with pytest.raises(SystemExit):
        with mock.patch('sys.stderr', new=output):
            critique._expand_directories(['not-a-schema.csv'])

    result = output.getvalue()
    assert 'ERROR: not-a-schema.csv is not a yaml or json file.' in result


def test_load_schemas_no_file():
    """Attempting to load a non-existent file should sys exit."""
    output = io.StringIO()

    with pytest.raises(SystemExit):
        with mock.patch('sys.stderr', new=output):
            critique._load_schemas('not-a-real-file.yaml')

    result = output.getvalue()
    assert 'ERROR: not-a-real-file.yaml does not exist.' in result


@mock.patch('horkos.cmdline.critique._assemble_catalog')
@mock.patch('horkos.cmdline.critique._load_schemas')
def test_relative_critique_already_exists(
        _load_schemas: mock.MagicMock,
        _assemble_catalog: mock.MagicMock,
):
    """Should error if the catalog already has a schema of the same name."""
    schema = mock.MagicMock()
    schema.name = 'foobar'
    _load_schemas.return_value = [schema]
    catalog = _assemble_catalog.return_value
    catalog.schemas = {'foobar': None}
    output = io.StringIO()

    with pytest.raises(SystemExit):
        with mock.patch('sys.stderr', new=output):
            critique._relative_critique('schema_file', ['catalog_files'])

    result = output.getvalue()
    assert 'A schema named foobar is already in the catalog' in result



import functools
import typing

from horkos import schemaomatic
from horkos import cataloger


def oxford_join(values: typing.List[str], last: str = 'and') -> str:
    """Join the given strings into an english list."""
    if not values:
        return ''
    if len(values) == 1:
        return values[0]
    working = values.copy()
    working[-1] = f'{last} {working[-1]}'
    joiner = ', ' if len(values) > 2 else ' '
    return joiner.join(working)


def schema_not_in_catalog(func: typing.Callable) -> typing.Callable:
    """
    Decorate a relative critique function to require a schema is not in
    the given catalog.
    """

    @functools.wraps(func)
    def wrapper(
            schema: schemaomatic.Schema,
            catalog: cataloger.Catalog,
            *args,
            **kwargs,
    ):
        """Require that the schema is not in the catalog."""
        if schema.name in catalog.schemas:
            raise ValueError(f'schema named {schema.name} already in catalog.')
        return func(schema, catalog, *args, **kwargs)
    return wrapper



import collections
import itertools
import typing

from horkos import cataloger
from horkos.critiquer import _definitions
from horkos.critiquer import _utils
from horkos import schemaomatic


@_utils.schema_not_in_catalog
def relative_type_consistency_check(
        schema: schemaomatic.Schema, catalog: cataloger.Catalog
) -> typing.List[_definitions.Critique]:
    """
    When comparing a schema against a catalog the schema's fields should have
    a typing consistent with the typing that already exists in the catalog.
    """
    fields = list(itertools.chain.from_iterable(
        schema.fields.values() for schema in catalog.schemas.values()
    ))
    type_map = collections.defaultdict(set)
    for field in fields:
        type_map[field.name].add(field.type_)
    critiques = []
    for field in schema.fields.values():
        if field.name not in type_map or field.type_ in type_map[field.name]:
            continue
        types = list(sorted(type_map[field.name]))
        types_str = _utils.oxford_join(types, last='or')
        msg = (
            f'{field.name} declared as {field.type_}, but other schemas in '
            f'the catalog have it declared as {types_str}. '
            f'The type of {field.name} should be consistent between schemas.'
        )
        critiques.append(_definitions.Critique(
            'relative', 'type_consistency', msg, schema.name, field.name
        ))
    return critiques

def catalog_type_consistency_check(
        catalog: cataloger.Catalog
) -> typing.List[_definitions.Critique]:
    """
    Field types should be consistent within the catalog.

    :param catalog:
        The group of schemas to check for type uniformity.
    :return:
        A list of _definitions.Critiques .
    """
    fields = list(itertools.chain.from_iterable(
        schema.fields.values() for schema in catalog.schemas.values()
    ))
    type_map = collections.defaultdict(set)
    for field in fields:
        type_map[field.name].add(field.type_)
    return [
        _definitions.Critique(
            'catalog',
            'type_consistency',
            (
                f'Catalog describes {field.name} as '
                f'{_utils.oxford_join(list(sorted(type_map[field.name])))}. '
                f'The type of {field.name} should be consistent between '
                'schemas.'
            ),
            schema.name,
            field.name
        )
        for schema in catalog.schemas.values()
        for field in schema.fields.values()
        if len(type_map[field.name]) > 1
    ]



import collections
import itertools
import re
import typing


from horkos import cataloger
from horkos.critiquer import _utils
from horkos.critiquer import _definitions
from horkos import schemaomatic


CAMEL_REGEX = re.compile(r'^[a-z]+([A-Z][a-z]*)*$')
CAPS_REGEX = re.compile(r'^([A-Z]+_{0,1})*[A-Z]+$')
PASCAL_REGEX = re.compile(r'^([A-Z][a-z]*)+$')
SNAKE_REGEX = re.compile(r'^([a-z]+_{0,1})*[a-z]+$')
CASE_MAP = {
    'CAPS_CASE': CAPS_REGEX,
    'camelCase': CAMEL_REGEX,
    'PascalCase': PASCAL_REGEX,
    'snake_case': SNAKE_REGEX,
}


def _casing_type(value: str) -> list:
    """Determine which casing types the value adheres to."""
    return [case for case, regex in CASE_MAP.items() if regex.match(value)]


def _most_common_casing_type(values: typing.List[str]) -> typing.Tuple:
    """Identify the most common casing type from the list of strings."""
    counter = collections.Counter(
        itertools.chain.from_iterable(
          _casing_type(value) for value in values
        )
    )
    common = counter.most_common()
    if not common:
        return [], 0
    count = common[0][1]
    casings = [c for c in CASE_MAP if counter[c] == count]
    return casings, count


def schema_uniform_field_casing_check(
        schema: schemaomatic.Schema
) -> typing.List[_definitions.Critique]:
    """Within a schema fields should have uniform casing."""
    casings, count = _most_common_casing_type(schema.fields.keys())
    if count == len(schema.fields):
        return []
    verb = 'is' if len(casings) == 1 else 'are'
    suggestions = _utils.oxford_join(casings)
    closest = (
        ''
        if count == 0
        else f' {suggestions} {verb} most common with {count} matching.'
    )
    msg = f'No clear casing convention.{closest}'
    return [_definitions.Critique(
        'schema', 'field_consistency', msg, schema.name
    )]


@_utils.schema_not_in_catalog
def relative_field_casing_check(
        schema: schemaomatic.Schema, catalog: cataloger.Catalog
) -> typing.List[_definitions.Critique]:
    """
    When comparing a schema against a catalog the schema should have
    casing conventions that are consistent with the catalog.
    """
    fields = list(itertools.chain.from_iterable(
        s.fields.keys() for s in catalog.schemas.values()
    ))
    catalog_casings, catalog_count = _most_common_casing_type(fields)
    percentage = int(catalog_count * 100 / (len(fields) or 1))
    conventions = _utils.oxford_join(catalog_casings, last='or')
    non_existing = set(schema.fields.keys()) - set(fields)
    critiques = []
    for field in non_existing:
        overlap = set(_casing_type(field)).intersection(catalog_casings)
        if overlap or not catalog_casings:
            continue
        msg = (
            f'The casing of {field} does not match the dominant '
            f'patterns of the catalog. {percentage}% of fields in the catalog '
            f'follow {conventions}.'
        )
        critiques.append(_definitions.Critique(
            'relative', 'field_consistency', msg, schema.name, field
        ))
    return critiques


def catalog_field_casing_check(
        catalog: cataloger.Catalog
) -> typing.List[_definitions.Critique]:
    """
    Naming conventions should be consistent across all fields in a catalog.

    :param catalog:
        The group of schemas to check for uniformity.
    :return:
        A _definitions.Critique if one is found, otherwise nothing.
    """
    fields = list(itertools.chain.from_iterable(
        s.fields.keys() for s in catalog.schemas.values()
    ))
    casings, count = _most_common_casing_type(fields)
    if len(fields) == count:
        return []
    verb = 'is' if len(casings) == 1 else 'are'
    suggestions = _utils.oxford_join(casings)
    percentage = int(count * 100 / (len(fields) or 1))
    closest = (
        ''
        if count == 0
        else f' {suggestions} {verb} most common with {percentage}% matching.'
    )
    msg = f'No clear casing convention.{closest}'
    return [_definitions.Critique('catalog', 'field_consistency', msg)]


def schema_prefer_snake_case(
        schema: schemaomatic.Schema
) -> typing.List[_definitions.Critique]:
    """
    snake_case should be the preferred.

    :param schema:
        The schema to validate for snake_case usage.
    :return:
        A _definitions.Critique if one is found, otherwise nothing.
    """
    casings, _ = _most_common_casing_type(schema.fields)
    if 'snake_case' in casings:
        return []
    joined = _utils.oxford_join(casings, last="or")
    current = (
        ''
        if not casings else
        f' Schema is currently using {joined}.'
    )
    msg = (
        'snake_case field names are recommended for the greatest '
        f'compatibility with common big data technologies.{current}'
    )
    return [
        _definitions.Critique(
            'schema', 'recommend_snake_case', msg, schema.name
        )
    ]


def catalog_prefer_snake_case(
        catalog: cataloger.Catalog
) -> typing.List[_definitions.Critique]:
    """
    snake_case should be the preferred.

    :param catalog:
        The catalog to validate for snake_case usage.
    :return:
        A _definitions.Critique if one is found, otherwise nothing.
    """
    fields = list(itertools.chain.from_iterable(
        schema.fields.keys() for schema in catalog.schemas.values()
    ))

    casings, _ = _most_common_casing_type(fields)
    if 'snake_case' in casings:
        return []
    joined = _utils.oxford_join(casings, last="or")
    current = (
        ''
        if not casings else
        f' Catalog is currently using {joined}.'
    )
    msg = (
        'snake_case field names are recommended for the greatest '
        f'compatibility with common big data technologies.{current}'
    )
    return [_definitions.Critique('catalog', 'recommend_snake_case', msg)]



import io
import os
from unittest import mock

import pytest

from horkos import cmdline

DIR_PATH = os.path.dirname(os.path.realpath(__file__))
RESOURCE_PATH = os.path.join(DIR_PATH, 'resources')


def test_check_cmdline():
    """Check cmdline should be able to run."""
    schema_file = os.path.join(RESOURCE_PATH, 'sample.yaml')
    data_file = os.path.join(RESOURCE_PATH, 'sample.csv')
    output = io.StringIO()

    with mock.patch('sys.stdout', new=output):
        cmdline.main(['check', '--schema', schema_file, data_file])

    result = output.getvalue()
    assert 'value of "ERROR" in method did not pass choice check' in result
    expected = 'value of "BAD" for response_code could not be cast to integer'
    assert expected in result
    assert '2 errors found' in result


def test_critique_cmdline_relative():
    """Critique cmdline should be able to run a relative comparison."""
    schema_file = os.path.join(RESOURCE_PATH, 'sample.yaml')
    output = io.StringIO()

    with pytest.raises(SystemExit):
        with mock.patch('sys.stdout', new=output):
            cmdline.main([
                'critique', '--schema', schema_file, '--catalog', RESOURCE_PATH
            ])

    result = output.getvalue()
    expected = '[field_consistency] http_request.response_code'
    assert expected in result
    expected = r'100% of fields in the catalog follow camelCase.'
    assert expected in result


def test_critique_cmdline_schema():
    """Critique cmdline should be able to run on a schema."""
    schema_file = os.path.join(RESOURCE_PATH, 'alternative.yaml')
    output = io.StringIO()

    with pytest.raises(SystemExit):
        with mock.patch('sys.stdout', new=output):
            cmdline.main(['critique', '--schema', schema_file])

    result = output.getvalue()
    expected = '[recommend_snake_case] business-card - snake_case field names'
    assert expected in result


def test_critique_cmdline_catalog():
    """Critique cmdline should be able to run a catalog critique."""
    output = io.StringIO()

    with pytest.raises(SystemExit):
        with mock.patch('sys.stdout', new=output):
            cmdline.main([
                'critique',
                '--catalog', RESOURCE_PATH,
                '--catalog', os.path.join(RESOURCE_PATH, 'sample.yaml')
            ])

    result = output.getvalue()
    expected = '[field_consistency]  - No clear casing convention.'
    assert expected in result


def test_critique_cmdline_catalog_no_flags():
    """Should get a bad error if we don't use the --schema or --catalog."""
    output = io.StringIO()

    with pytest.raises(SystemExit):
        with mock.patch('sys.stderr', new=output):
            cmdline.main(['critique'])

    result = output.getvalue()
    expected = '--schema or --catalog must be specified'
    assert expected in result


def test_critique_cmdline_schema_in_catalog_files():
    """Critique cmdline should give a sys exit if schema file is in catalog."""
    schema_file = os.path.join(RESOURCE_PATH, 'alternative.yaml')
    output = io.StringIO()

    with pytest.raises(SystemExit):
        with mock.patch('sys.stderr', new=output):
            cmdline.main([
                'critique', '--schema', schema_file, '--catalog', schema_file
            ])

    result = output.getvalue()
    expected = 'is given for both --schema and --catalog'
    assert expected in result



import argparse

from horkos.cmdline import check
from horkos.cmdline import critique

_HANDLERS = {
    'check': check,
    'critique': critique,
}

def _parse(args: list = None) -> dict:
    """Parse the given arguments and return them as a simple dictionary."""
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest='subcommand', required=True)
    for name, handler in _HANDLERS.items():
        subparser = subparsers.add_parser(name, help=handler.__doc__)
        handler.configure_parser(subparser)
    return vars(parser.parse_args(args))


def main(args: list = None):
    """Run the horkos command."""
    parsed_args = _parse(args)
    command = parsed_args.pop('subcommand')
    _HANDLERS[command].main(**parsed_args)



from unittest import mock

from horkos.critiquer import _typing


def test_catalog_type_consistency_check_with_critique():
    """The types of each field should be consistent across schemas."""
    field_1 = mock.MagicMock()
    field_1.name = 'foobar'
    field_1.type_ = 'string'
    field_2 = mock.MagicMock()
    field_2.name = 'foobar'
    field_2.type_ = 'integer'
    catalog = mock.MagicMock(
        schemas={
            'ds1': mock.MagicMock(fields={'foobar': field_1}),
            'ds2': mock.MagicMock(fields={'foobar': field_2}),
        }
    )

    result = _typing.catalog_type_consistency_check(catalog)

    assert result
    expected = (
        'Catalog describes foobar as integer and string. The type of foobar '
        'should be consistent between schemas.'
    )
    assert expected in result[0].message


def test_catalog_type_consistency_check_without_critique():
    """If types are consistent then no critique should be returned."""
    field_1 = mock.MagicMock()
    field_1.name = 'foobar'
    field_1.type_ = 'integer'
    field_2 = mock.MagicMock()
    field_2.name = 'foobar'
    field_2.type_ = 'integer'
    catalog = mock.MagicMock(
        schemas={
            'ds1': mock.MagicMock(fields={'foobar': field_1}),
            'ds2': mock.MagicMock(fields={'foobar': field_2}),
        }
    )

    result = _typing.catalog_type_consistency_check(catalog)

    assert not result


def test_relative_type_consistency_check_with_critique():
    """The types of a new schema should be consistent with existing"""
    existing_field = mock.MagicMock()
    existing_field.name = 'foobar'
    existing_field.type_ = 'integer'
    existing_schema = mock.MagicMock()
    existing_schema.fields = {'foobar': existing_field}
    catalog = mock.MagicMock(schemas={'ds1': existing_schema})
    new_field = mock.MagicMock()
    new_field.name = 'foobar'
    new_field.type_ = 'string'
    new_schema = mock.MagicMock()
    new_schema.fields = {'foobar': new_field}

    result = _typing.relative_type_consistency_check(new_schema, catalog)

    assert result
    expected = 'other schemas in the catalog have it declared as integer'
    assert expected in result[0].message


def test_relative_type_consistency_check_without_critique():
    """If types are consistent, then no critiques should be raised."""
    existing_field = mock.MagicMock()
    existing_field.name = 'foobar'
    existing_field.type_ = 'integer'
    existing_schema = mock.MagicMock()
    existing_schema.fields = {'foobar': existing_field}
    catalog = mock.MagicMock(schemas={'ds1': existing_schema})
    new_field = mock.MagicMock()
    new_field.name = 'foobar'
    new_field.type_ = 'integer'
    new_schema = mock.MagicMock()
    new_schema.fields = {'foobar': new_field}

    result = _typing.relative_type_consistency_check(new_schema, catalog)

    assert not result



import typing

from horkos import cataloger
from horkos import schemaomatic
from horkos.critiquer._definitions import Critique
from horkos.critiquer import _casing
from horkos.critiquer import _typing


_CATALOG_CRITIQUERS = [
    _casing.catalog_prefer_snake_case,
    _casing.catalog_field_casing_check,
    _typing.catalog_type_consistency_check,
]
_RELATIVE_CRITIQUERS = [
    _casing.relative_field_casing_check,
    _typing.relative_type_consistency_check,
]
_SCHEMA_CRITIQUERS = [
    _casing.schema_prefer_snake_case,
    _casing.schema_uniform_field_casing_check,
]


def critique_addition(
        schema: schemaomatic.Schema,
        catalog: cataloger.Catalog,
) -> typing.List[Critique]:
    """
    Critique a schema within the context of adding it to the given catalog
    of schemas.

    :param schema:
        The schema to critique.
    :param catalog:
        The catalog of schemas it is being added to.
    :return:
        A list of critiques of the schema in the context of the catalog.
    """
    result = []
    for critiquer in _RELATIVE_CRITIQUERS:
        result.extend(critiquer(schema, catalog))
    return result


def critique_catalog(catalog: cataloger.Catalog) -> typing.List[Critique]:
    """
    Critique the schemas within the catalog as a cohesive group.

    :param catalog:
        The catalog to critique.
    :return:
        A list of critiques of the catalog.
    """
    result = []
    for critiquer in _CATALOG_CRITIQUERS:
        result.extend(critiquer(catalog))
    return result


def critique_schema(schema: schemaomatic.Schema) -> typing.List[Critique]:
    """
    Critique a specific schema for self contained issues.

    :param schema:
        The schema to critique.
    :return:
        A list of critiques of the schema.
    """
    result = []
    for critiquer in _SCHEMA_CRITIQUERS:
        result.extend(critiquer(schema))
    return result


__all__ = ('critique_addition', 'critique_catalog', 'critique_schema')



import typing


class Critique(typing.NamedTuple):
    """A critique of a dataschema."""

    scope: str
    code: str
    message: str
    schema_name: str = None
    field_name: str = None

    def __str__(self):
        """Convert the critique to a string."""
        parts = [self.schema_name, self.field_name]
        parts = [p for p in parts if p is not None]
        name = '.'.join(parts)
        return f'[{self.code}] {name} - {self.message}'



"""
Critique a schema or catalog of schemas, identifying common mistakes or
potential issues.
"""
import argparse
import os
from os import path
import re
import sys
import typing

import yaml

from horkos import errors
from horkos import cataloger
from horkos import critiquer
from horkos import schemaomatic

SCHEMA_FILE_RE = re.compile(r'.*\.(yaml|yml|json)$')


def configure_parser(parser: argparse.ArgumentParser):
    """Configure the given parser to work for the critique command."""
    parser.add_argument(
        '--schema', '-s',
        dest='schema_file',
        help='The schema file to critique.'
    )
    parser.add_argument(
        '--catalog', '-c',
        action='append',
        dest='catalog_files',
        help=(
            'A path defining the catalog of schemas. If a directory is given '
            'all yaml files contained within will be used as schemas in the '
            'catalog. If used without the `--schema` flag the catalog itself '
            'will be critiqued, otherwise the flagged schema will be '
            'critiqued within the context of the given catalog.'
        )
    )


def _expand_directories(file_paths: typing.List[str]) -> typing.List[str]:
    """
    Expand any directories in the given set of paths replacing them with all
    of the yaml files they contain.
    """
    result = []
    for file_path in file_paths:
        if not path.exists(file_path):
            print(f'ERROR: {file_path} does not exist.', file=sys.stderr)
            sys.exit(1)
        if not path.isdir(file_path):
            if not SCHEMA_FILE_RE.match(file_path):
                print(
                    f'ERROR: {file_path} is not a yaml or json file.',
                    file=sys.stderr
                )
                sys.exit(1)
            result.append(file_path)
            continue
        for root, _, files in os.walk(file_path):
            result.extend(
                path.abspath(path.join(root, f))
                for f in files
                if SCHEMA_FILE_RE.match(f)
            )
    return result


def _load_schemas(schema_file: str) -> typing.List[schemaomatic.Schema]:
    """Load the given schema file returnings all schemas it contains."""
    if not path.isfile(schema_file):
        print(f'ERROR: {schema_file} does not exist.', file=sys.stderr)
        sys.exit(1)
    with open(schema_file) as f:
        contents = f.read()
    return [
        schemaomatic.load_schema(doc) for doc in yaml.safe_load_all(contents)
    ]


def _assemble_catalog(catalog_files: typing.List[str], omit: str = None):
    """Assemble a catalog using the given catalog files."""
    if omit in catalog_files:
        print(
            f'ERROR: {omit} is given for both --schema and --catalog',
            file=sys.stderr,
        )
        sys.exit(1)
    expanded_catalog = set(_expand_directories(catalog_files))
    if omit:
        expanded_catalog -= {path.abspath(omit)}
    catalog = cataloger.Catalog()
    for file_path in expanded_catalog:
        with open(file_path) as f:
            contents = f.read()
        for doc in yaml.safe_load_all(contents):
            catalog.update(doc)
    return catalog


def _schema_critique(schema_file: str) -> typing.List[critiquer.Critique]:
    """Critique a schema."""
    critiques = []
    for schema in _load_schemas(schema_file):
        critiques.extend(critiquer.critique_schema(schema))
    return critiques


def _relative_critique(
        schema_file: str, catalog_files: typing.List[str]
) -> typing.List[critiquer.Critique]:
    """Critique the schema within the context of the catalog."""
    schemas = _load_schemas(schema_file)
    catalog = _assemble_catalog(catalog_files, omit=schema_file)
    for schema in schemas:
        if schema.name not in catalog.schemas:
            continue
        print(
            f'A schema named {schema.name} is already in the catalog',
            file=sys.stderr,
        )
        sys.exit(1)
    critiques = []
    for schema in _load_schemas(schema_file):
        critiques.extend(critiquer.critique_addition(schema, catalog))
    return critiques


def _catalog_critique(
        catalog_files: typing.List[str]
) -> typing.List[critiquer.Critique]:
    """Critique the catalog."""
    catalog = _assemble_catalog(catalog_files)
    return critiquer.critique_catalog(catalog)


def main(
        schema_file: str,
        catalog_files: typing.List[str],
        **kwargs,
):
    """Critique the given schema / catalog."""
    if not schema_file and not catalog_files:
        print('--schema or --catalog must be specified', file=sys.stderr)
        sys.exit(1)
    if schema_file and not catalog_files:
        critiques = _schema_critique(schema_file)
    if schema_file and catalog_files:
        critiques = _relative_critique(schema_file, catalog_files)
    if not schema_file and catalog_files:
        critiques = _catalog_critique(catalog_files)
    for critique in critiques:
        print(critique)
    if critiques:
        sys.exit(1)
